{-# OPTIONS -fglasgow-exts -fallow-undecidable-instances #-}

-- Arithmetics on arbitrary-precision decimal TYPES

module ArbArithmT 
   (DigitsInReverse, NoLeadingZeros, FullAdder, Carry0, Carry1,
    CardSum,
    vappend, vhead, vtail
   ) where

import ArbPrecDecT

--import Data.Array

-- Invert the sequence of digits

class DigitsInReverse' df w dr | df w -> dr

instance DigitsInReverse' Sz acc acc

instance (Digits (d drest), DigitsInReverse' drest (d acc) dr) 
	  => DigitsInReverse' (d drest) acc dr

class DigitsInReverse df dr | df -> dr, dr -> df

instance (DigitsInReverse' df Sz dr, DigitsInReverse' dr Sz df) => 
	 DigitsInReverse df dr

digits_rev:: (Digits ds, Digits dsr, DigitsInReverse ds dsr) => ds -> dsr
digits_rev = undefined

-- *ArbArithmT> :t digits_rev (D1$D2$D3 Sz)
-- *ArbArithmT> :t (\v -> digits_rev v `asTypeOf` (D1$D2$D3 Sz))
{-

class ZeroPad' d1 d2 d1o | d1 d2 -> d1o

instance ZeroPad' d1 Sz d1
instance (ZeroPad' Sz d2r d1r') 
	 => ZeroPad' Sz (d2 d2r) (D0 d1r')
instance (ZeroPad' d1r d2r d1r')
         => ZeroPad' (d1 d1r) (d2 d2r) (d1 d1r')
-}

-- The sequence d0 is d without leading zeros
class NoLeadingZeros d d0 | d -> d0
instance NoLeadingZeros Sz Sz
instance (NoLeadingZeros d d') => NoLeadingZeros (D0 d) d'
instance NoLeadingZeros (D1 d) (D1 d)
instance NoLeadingZeros (D2 d) (D2 d)
instance NoLeadingZeros (D3 d) (D3 d)
instance NoLeadingZeros (D4 d) (D4 d)
instance NoLeadingZeros (D5 d) (D5 d)
instance NoLeadingZeros (D6 d) (D6 d)
instance NoLeadingZeros (D7 d) (D7 d)
instance NoLeadingZeros (D8 d) (D8 d)
instance NoLeadingZeros (D9 d) (D9 d)


--class ZeroPad d1 d2 d1o | d1 d2 -> d1o, d1o -> d1
--class ZeroPad d1 d2 d1o d2o | d1 d2 -> d1o, d1 d2 -> d2o, d1o -> d1, d2o -> d2, d1 d2o -> d2
--class ZeroPad d1 d2 d1o d2o | d1 d2 -> d1o d2o, d1o d2o -> d1 d2

{-
instance (ZeroPad' d1 d2 d1o,
          DigitsInReverse d1o d1o', NoLeadingZeros d1o' d1', 
	  DigitsInReverse d1' d1)
	 => ZeroPad d1 d2 d1o
-}


-- *ArbArithmT> :t test_pad (D1$D2 Sz) (D1$D3 Sz)
-- D1 (D2 Sz)
-- *ArbArithmT> :t test_pad (D1$D2 Sz) (D1$D3$D4 Sz)
-- D1 (D2 (D0 Sz))

-- test_pad:: (Digits ds1, Digits ds2, Digits ds1r,
--             ZeroPad ds1 ds2 ds1r) => ds1 -> ds2 -> ds1r
-- test_pad = undefined


-- d1 + d2 + carry_in -> (sum + carry_out)
class FullAdder d1 d2 cin dr cout
      | d1 d2 cin -> cout, d1 d2 cin -> dr, 
        d1 dr cin -> cout, d1 dr cin -> d2 
  where
    _unused:: (d1 xd1) -> (d2 xd2) -> cin -> (dr xdr)
    _unused = undefined

data Carry0
data Carry1

instance FullAdder D0 D0 Carry0 D0 Carry0
instance FullAdder D0 D0 Carry1 D1 Carry0
instance FullAdder D0 D1 Carry0 D1 Carry0
instance FullAdder D0 D1 Carry1 D2 Carry0
instance FullAdder D0 D2 Carry0 D2 Carry0
instance FullAdder D0 D2 Carry1 D3 Carry0
instance FullAdder D0 D3 Carry0 D3 Carry0
instance FullAdder D0 D3 Carry1 D4 Carry0
instance FullAdder D0 D4 Carry0 D4 Carry0
instance FullAdder D0 D4 Carry1 D5 Carry0
instance FullAdder D0 D5 Carry0 D5 Carry0
instance FullAdder D0 D5 Carry1 D6 Carry0
instance FullAdder D0 D6 Carry0 D6 Carry0
instance FullAdder D0 D6 Carry1 D7 Carry0
instance FullAdder D0 D7 Carry0 D7 Carry0
instance FullAdder D0 D7 Carry1 D8 Carry0
instance FullAdder D0 D8 Carry0 D8 Carry0
instance FullAdder D0 D8 Carry1 D9 Carry0
instance FullAdder D0 D9 Carry0 D9 Carry0
instance FullAdder D0 D9 Carry1 D0 Carry1
instance FullAdder D1 D0 Carry0 D1 Carry0
instance FullAdder D1 D0 Carry1 D2 Carry0
instance FullAdder D1 D1 Carry0 D2 Carry0
instance FullAdder D1 D1 Carry1 D3 Carry0
instance FullAdder D1 D2 Carry0 D3 Carry0
instance FullAdder D1 D2 Carry1 D4 Carry0
instance FullAdder D1 D3 Carry0 D4 Carry0
instance FullAdder D1 D3 Carry1 D5 Carry0
instance FullAdder D1 D4 Carry0 D5 Carry0
instance FullAdder D1 D4 Carry1 D6 Carry0
instance FullAdder D1 D5 Carry0 D6 Carry0
instance FullAdder D1 D5 Carry1 D7 Carry0
instance FullAdder D1 D6 Carry0 D7 Carry0
instance FullAdder D1 D6 Carry1 D8 Carry0
instance FullAdder D1 D7 Carry0 D8 Carry0
instance FullAdder D1 D7 Carry1 D9 Carry0
instance FullAdder D1 D8 Carry0 D9 Carry0
instance FullAdder D1 D8 Carry1 D0 Carry1
instance FullAdder D1 D9 Carry0 D0 Carry1
instance FullAdder D1 D9 Carry1 D1 Carry1
instance FullAdder D2 D0 Carry0 D2 Carry0
instance FullAdder D2 D0 Carry1 D3 Carry0
instance FullAdder D2 D1 Carry0 D3 Carry0
instance FullAdder D2 D1 Carry1 D4 Carry0
--
instance FullAdder D2 D2 Carry0 D4 Carry0
instance FullAdder D2 D2 Carry1 D5 Carry0
instance FullAdder D2 D3 Carry0 D5 Carry0
instance FullAdder D2 D3 Carry1 D6 Carry0
instance FullAdder D2 D4 Carry0 D6 Carry0
instance FullAdder D2 D4 Carry1 D7 Carry0
instance FullAdder D2 D5 Carry0 D7 Carry0
instance FullAdder D2 D5 Carry1 D8 Carry0
instance FullAdder D2 D6 Carry0 D8 Carry0
instance FullAdder D2 D6 Carry1 D9 Carry0
instance FullAdder D2 D7 Carry0 D9 Carry0
instance FullAdder D2 D7 Carry1 D0 Carry1
instance FullAdder D2 D8 Carry0 D0 Carry1
instance FullAdder D2 D8 Carry1 D1 Carry1
instance FullAdder D2 D9 Carry0 D1 Carry1
instance FullAdder D2 D9 Carry1 D2 Carry1
instance FullAdder D3 D0 Carry0 D3 Carry0
instance FullAdder D3 D0 Carry1 D4 Carry0
instance FullAdder D3 D1 Carry0 D4 Carry0
instance FullAdder D3 D1 Carry1 D5 Carry0
instance FullAdder D3 D2 Carry0 D5 Carry0
instance FullAdder D3 D2 Carry1 D6 Carry0
instance FullAdder D3 D3 Carry0 D6 Carry0
instance FullAdder D3 D3 Carry1 D7 Carry0
instance FullAdder D3 D4 Carry0 D7 Carry0
instance FullAdder D3 D4 Carry1 D8 Carry0
instance FullAdder D3 D5 Carry0 D8 Carry0
instance FullAdder D3 D5 Carry1 D9 Carry0
instance FullAdder D3 D6 Carry0 D9 Carry0
instance FullAdder D3 D6 Carry1 D0 Carry1
instance FullAdder D3 D7 Carry0 D0 Carry1
instance FullAdder D3 D7 Carry1 D1 Carry1
instance FullAdder D3 D8 Carry0 D1 Carry1
instance FullAdder D3 D8 Carry1 D2 Carry1
instance FullAdder D3 D9 Carry0 D2 Carry1
instance FullAdder D3 D9 Carry1 D3 Carry1
instance FullAdder D4 D0 Carry0 D4 Carry0
instance FullAdder D4 D0 Carry1 D5 Carry0
instance FullAdder D4 D1 Carry0 D5 Carry0
instance FullAdder D4 D1 Carry1 D6 Carry0
instance FullAdder D4 D2 Carry0 D6 Carry0
instance FullAdder D4 D2 Carry1 D7 Carry0
instance FullAdder D4 D3 Carry0 D7 Carry0
instance FullAdder D4 D3 Carry1 D8 Carry0
instance FullAdder D4 D4 Carry0 D8 Carry0
instance FullAdder D4 D4 Carry1 D9 Carry0
instance FullAdder D4 D5 Carry0 D9 Carry0
instance FullAdder D4 D5 Carry1 D0 Carry1
instance FullAdder D4 D6 Carry0 D0 Carry1
instance FullAdder D4 D6 Carry1 D1 Carry1
instance FullAdder D4 D7 Carry0 D1 Carry1
instance FullAdder D4 D7 Carry1 D2 Carry1
instance FullAdder D4 D8 Carry0 D2 Carry1
instance FullAdder D4 D8 Carry1 D3 Carry1
instance FullAdder D4 D9 Carry0 D3 Carry1
instance FullAdder D4 D9 Carry1 D4 Carry1
instance FullAdder D5 D0 Carry0 D5 Carry0
instance FullAdder D5 D0 Carry1 D6 Carry0
instance FullAdder D5 D1 Carry0 D6 Carry0
instance FullAdder D5 D1 Carry1 D7 Carry0
instance FullAdder D5 D2 Carry0 D7 Carry0
instance FullAdder D5 D2 Carry1 D8 Carry0
instance FullAdder D5 D3 Carry0 D8 Carry0
instance FullAdder D5 D3 Carry1 D9 Carry0
instance FullAdder D5 D4 Carry0 D9 Carry0
instance FullAdder D5 D4 Carry1 D0 Carry1
instance FullAdder D5 D5 Carry0 D0 Carry1
instance FullAdder D5 D5 Carry1 D1 Carry1
instance FullAdder D5 D6 Carry0 D1 Carry1
instance FullAdder D5 D6 Carry1 D2 Carry1
instance FullAdder D5 D7 Carry0 D2 Carry1
instance FullAdder D5 D7 Carry1 D3 Carry1
instance FullAdder D5 D8 Carry0 D3 Carry1
instance FullAdder D5 D8 Carry1 D4 Carry1
instance FullAdder D5 D9 Carry0 D4 Carry1
instance FullAdder D5 D9 Carry1 D5 Carry1
instance FullAdder D6 D0 Carry0 D6 Carry0
instance FullAdder D6 D0 Carry1 D7 Carry0
instance FullAdder D6 D1 Carry0 D7 Carry0
instance FullAdder D6 D1 Carry1 D8 Carry0
instance FullAdder D6 D2 Carry0 D8 Carry0
instance FullAdder D6 D2 Carry1 D9 Carry0
instance FullAdder D6 D3 Carry0 D9 Carry0
instance FullAdder D6 D3 Carry1 D0 Carry1
instance FullAdder D6 D4 Carry0 D0 Carry1
instance FullAdder D6 D4 Carry1 D1 Carry1
instance FullAdder D6 D5 Carry0 D1 Carry1
instance FullAdder D6 D5 Carry1 D2 Carry1
instance FullAdder D6 D6 Carry0 D2 Carry1
instance FullAdder D6 D6 Carry1 D3 Carry1
instance FullAdder D6 D7 Carry0 D3 Carry1
instance FullAdder D6 D7 Carry1 D4 Carry1
instance FullAdder D6 D8 Carry0 D4 Carry1
instance FullAdder D6 D8 Carry1 D5 Carry1
instance FullAdder D6 D9 Carry0 D5 Carry1
instance FullAdder D6 D9 Carry1 D6 Carry1
instance FullAdder D7 D0 Carry0 D7 Carry0
instance FullAdder D7 D0 Carry1 D8 Carry0
instance FullAdder D7 D1 Carry0 D8 Carry0
instance FullAdder D7 D1 Carry1 D9 Carry0
instance FullAdder D7 D2 Carry0 D9 Carry0
instance FullAdder D7 D2 Carry1 D0 Carry1
instance FullAdder D7 D3 Carry0 D0 Carry1
instance FullAdder D7 D3 Carry1 D1 Carry1
instance FullAdder D7 D4 Carry0 D1 Carry1
instance FullAdder D7 D4 Carry1 D2 Carry1
instance FullAdder D7 D5 Carry0 D2 Carry1
instance FullAdder D7 D5 Carry1 D3 Carry1
instance FullAdder D7 D6 Carry0 D3 Carry1
instance FullAdder D7 D6 Carry1 D4 Carry1
instance FullAdder D7 D7 Carry0 D4 Carry1
instance FullAdder D7 D7 Carry1 D5 Carry1
instance FullAdder D7 D8 Carry0 D5 Carry1
instance FullAdder D7 D8 Carry1 D6 Carry1
instance FullAdder D7 D9 Carry0 D6 Carry1
instance FullAdder D7 D9 Carry1 D7 Carry1
instance FullAdder D8 D0 Carry0 D8 Carry0
instance FullAdder D8 D0 Carry1 D9 Carry0
instance FullAdder D8 D1 Carry0 D9 Carry0
instance FullAdder D8 D1 Carry1 D0 Carry1
instance FullAdder D8 D2 Carry0 D0 Carry1
instance FullAdder D8 D2 Carry1 D1 Carry1
instance FullAdder D8 D3 Carry0 D1 Carry1
instance FullAdder D8 D3 Carry1 D2 Carry1
instance FullAdder D8 D4 Carry0 D2 Carry1
instance FullAdder D8 D4 Carry1 D3 Carry1
instance FullAdder D8 D5 Carry0 D3 Carry1
instance FullAdder D8 D5 Carry1 D4 Carry1
instance FullAdder D8 D6 Carry0 D4 Carry1
instance FullAdder D8 D6 Carry1 D5 Carry1
instance FullAdder D8 D7 Carry0 D5 Carry1
instance FullAdder D8 D7 Carry1 D6 Carry1
instance FullAdder D8 D8 Carry0 D6 Carry1
instance FullAdder D8 D8 Carry1 D7 Carry1
instance FullAdder D8 D9 Carry0 D7 Carry1
--
instance FullAdder D8 D9 Carry1 D8 Carry1
instance FullAdder D9 D0 Carry0 D9 Carry0
instance FullAdder D9 D0 Carry1 D0 Carry1
instance FullAdder D9 D1 Carry0 D0 Carry1
instance FullAdder D9 D1 Carry1 D1 Carry1
instance FullAdder D9 D2 Carry0 D1 Carry1
instance FullAdder D9 D2 Carry1 D2 Carry1
instance FullAdder D9 D3 Carry0 D2 Carry1
instance FullAdder D9 D3 Carry1 D3 Carry1
instance FullAdder D9 D4 Carry0 D3 Carry1
instance FullAdder D9 D4 Carry1 D4 Carry1
instance FullAdder D9 D5 Carry0 D4 Carry1
instance FullAdder D9 D5 Carry1 D5 Carry1
instance FullAdder D9 D6 Carry0 D5 Carry1
instance FullAdder D9 D6 Carry1 D6 Carry1
instance FullAdder D9 D7 Carry0 D6 Carry1
instance FullAdder D9 D7 Carry1 D7 Carry1
instance FullAdder D9 D8 Carry0 D7 Carry1
instance FullAdder D9 D8 Carry1 D8 Carry1
instance FullAdder D9 D9 Carry0 D8 Carry1
instance FullAdder D9 D9 Carry1 D9 Carry1
-- Could be used in template Haskell
-- Currently we just used cutting and pasting from an Emacs buffer
-- with GHCi into the emacs buffer with the code.
make_full_adder 
    = mapM_ putStrLn 
            [unwords $ doit d1 d2 cin | d1<-[0..9], d2<-[0..9], cin<-[0..1]]
  where
       doit d1 d2 cin 
          = ["instance FullAdder", tod d1, tod d2, toc cin, tod d12, toc cout]
	 where 
	     (d12,cout) = let sum = d1 + d2 + cin
	                  in if sum >= 10 then (sum-10,1) else (sum,0)
       tod n | (n >= 0 && 9 >= n) = "D" ++ (show n)
       toc 0 = "Carry0"
       toc 1 = "Carry1"
       

-- drs = ds1 + ds2 + carry_in
class DigitsSum ds1 ds2 cin dsr | ds1 ds2 cin -> dsr
instance DigitsSum Sz Sz Carry0 Sz
instance DigitsSum Sz Sz Carry1 (D1 Sz)
instance (DigitsSum (D0 Sz) (d2 d2rest) cin (d12 d12rest)) =>
         DigitsSum Sz (d2 d2rest) cin (d12 d12rest)
instance (DigitsSum (d1 d1rest) (D0 Sz) cin (d12 d12rest)) =>
         DigitsSum (d1 d1rest) Sz cin (d12 d12rest)

instance (FullAdder d1 d2 cin d12 cout, 
          DigitsSum d1rest d2rest cout d12rest) =>
   DigitsSum (d1 d1rest) (d2 d2rest) cin (d12 d12rest)

-- drs = ds1 + ds2 + carry_in
-- But now drs is known but ds2 is not
-- carry_in actually means borrow
-- Because our numbers are all non-negative, dsr should have
-- at least as many digits as ds1. Otherwise, the relation is
-- undefined and the subtraction is not possible.
class DigitsDif ds1 ds2 cin dsr | ds1 dsr cin -> ds2
instance DigitsDif Sz ds Carry0 ds
--instance DigitsDif' Sz Sz Carry1 (D1 Sz)
instance (DigitsDif (D0 Sz) (d2 d2rest) Carry1 (d12 d12rest)) =>
         DigitsDif Sz (d2 d2rest) Carry1 (d12 d12rest)

instance (FullAdder d1 d2 cin d12 cout, 
          DigitsDif d1rest d2rest cout d12rest) =>
   DigitsDif (d1 d1rest) (d2 d2rest) cin (d12 d12rest)

-- c12= c1 + c2
class (Card c1, Card c2, Card c12) => 
    CardSum c1 c2 c12 | c1 c2 -> c12, c1 c12 -> c2

instance (Card c1, Card c2, Card c12,
	  DigitsInReverse c1 c1r, 
          DigitsInReverse c2 c2r,
	  DigitsSum c1r c2r Carry0 c12r,
	  DigitsDif c1r c2r' Carry0 c12r,
	  DigitsInReverse c2r' c2', NoLeadingZeros c2' c2,
	  DigitsInReverse c12r c12)
	  => CardSum c1 c2 c12
	  
card_sum:: CardSum c1 c2 c12 => c1 -> c2 -> c12
card_sum = undefined

-- Tests of addition and subtraction

--class CardSum1 c1 c2 c12 | c1 c2 -> c12
-- :t card_sum (D1$ D0$ D1 Sz) (D1 Sz)
-- :t card_sum (D1 Sz) (D1$ D0$ D1 Sz) 
-- :t card_sum (D1$D9 Sz) (D1 Sz)
-- :t card_sum (D1 Sz) (D1$D9 Sz)
-- :t card_sum (D1$D1 Sz) (D1$D9 Sz)
-- :t card_sum (D9$D9 Sz) (D3 Sz)
-- :t card_sum (D1 Sz) (D9$D9 Sz)
-- :t \v -> card_sum (D1$D1 Sz) v `asTypeOf` (D3$D0 Sz)
-- :t \v -> card_sum (D1 Sz) v `asTypeOf` (D1$D0$D0 Sz)
-- :t \v -> card_sum (D9$D9 Sz) v `asTypeOf` (D1$D0$D0 Sz)

-- We could have used listVec' here

vappend va vb = 
  listVec (card_sum (vlength_t va) (vlength_t vb)) $ (velems va) ++ (velems vb)

-- The inferred type of vappend is
-- *ArbArithmT> :t vappend
-- vappend :: (CardSum size size1 c12) =>
-- 	   Vec size a -> Vec size1 a -> Vec c12 a
-- It could not have been clearer.


-- Tests of using vappend

{--
:t vappend v13i v13i
:t vappend (vec (D9 Sz) 0) (vec (D1 Sz) 0)
:t vappend (vec (D9$D9 Sz) 0) (vec (D4 Sz) 0)
:t vappend (vec (D9$D9$D9 Sz) 0) (vec (D4 Sz) 0)
:t vappend (vec (D9$D9$D9 Sz) 0) (vec (D1 Sz) 0)
:t vappend (vec (D1 Sz) 0) (vec (D9$D9$D9 Sz) 0) 
:t vappend (vec (D2$D5 Sz) 0) (vec (D9$D7$D9 Sz) 0) 
vappend (vec (D1$D0 Sz) 1) (vec (D1 Sz) 2)

--}

vhead:: CardSum (D1 Sz) size1 size => Vec size a -> Vec (D1 Sz) a
vhead va = listVec (D1 Sz) $ [head (velems va)]

-- We could have used listVec' here

vtail:: CardSum (D1 Sz) size1 size => Vec size a -> Vec size1 a
vtail va = result
  where result = listVec (vlength_t result) $ tail (velems va)
	
--
testc1 =
  let va = vec (D1$D2 Sz) 0
      vb = vec (D5 Sz) 1
      vc = vec (D8 Sz) 2
  in vzipWith (+) va (vappend vb (vtail vc))
--

-- The following leads to a type error, as expected:
{-
testc1_bad =
  let va = vec (D1$D2 Sz) 0
      vb = vec (D5 Sz) 1
      vc = vec (D9 Sz) 2
  in vzipWith (+) va (vappend vb (vtail vc))
-}

-- A simple vreverse, an example of using number-parameterized vectors
-- in a dynamic context

vreverse v = listVec (vlength_t v) $ reverse $ velems v

-- *ArbArithmT> :t vreverse
-- vreverse :: (Card size) => Vec size a -> Vec size a
-- The use of listVec implies a dynamic test -- as a witness to `acquire'
-- the static type `size' (the same type as the input vector)

testv = let v  = vappend (vec (D3 Sz) 1) (vec (D1 Sz) 4)
	    vr = vreverse v
	in vhead (vtail (vtail vr))
