{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}
{-# OPTIONS -fallow-overlapping-instances #-}

-- Number-parameterized typed in a dynamic context

module Reverse where

import ArbArithmT hiding (vhead, vtail, vappend, CardSum)
import ArbPrecDecT 
	(D0(..), D1(..), D2(..), D3(..), D4(..), D5(..), D6(..), 
	 D7(..), D8(..), D9(..), Sz(..), 
	 Digits(..), Card(..), CardC, cardc)

import Data.Array

{--
 An example from Hongwei Xi's site:
http://www-2.cs.cmu.edu/~hwxi/DML/examples/
 
(*
 * This simple example says that reversing a
 * list yields a list of the same length.
 *)

fun('a) reverse(l) =
let
    fun aux(nil, ys) = ys
      | aux(x::xs, ys) = aux(xs, x::ys)
    where aux <| {m:nat,n:nat} 'a list(m) * 'a list(n) -> 'a list(m+n)
in
    aux(l, [])
end
where reverse <| {n:nat} 'a list(n) -> 'a list(n)

We must note one dynamic test per iteration: pattern-match in aux.
--}

newtype Vec size a = Vec (Array Int a) deriving Show

vec:: (Card size) => size -> a -> Vec size a
vec size elem = Vec $ listArray (0,(c2num size)-1) $ repeat elem

listVec:: (Card size) => size -> [a] -> Vec size a
listVec size elems = Vec $ listArray (0,(c2num size)-1) elems

vlength_t::(Card size) => Vec size a -> size
vlength_t _ = undefined

vlength:: Vec size a -> Int
vlength (Vec arr) = let (0,last) = bounds arr in last+1

vat::  Vec size a -> Int -> a
vat (Vec arr) i = arr ! i

velems:: Vec size a -> [a]
velems (Vec v) = elems v

vzipWith:: (Card size) => 
           (a->b->c) -> (Vec size a) -> (Vec size b) -> (Vec size c)
vzipWith f va@(Vec a) (Vec b) = 
   listVec (vlength_t va) $
       zipWith f (elems a) (elems b)

v12c = listVec (D1 $ D2 Sz) ['a'..'z']
v12i = listVec (D1 $ D2 Sz) [1..]
v13i = listVec (D1 $ D3 Sz) [1..]

test1 = vlength v12c; test1' = vlength_t v12c
test2 = [vat v12c 0, vat v12c 1, vat v12c 2]

test3 = vzipWith (\c i -> (toEnum $ fromEnum c + fromInteger i)::Char) v12c v12i
test4 = vzipWith (+) v12i v12i

digits_rev:: (Digits ds, Digits dsr, DigitsInReverse ds dsr) => ds -> dsr
digits_rev = undefined

vappend va vb = 
  listVec (card_sum (vlength_t va) (vlength_t vb)) $ (velems va) ++ (velems vb)

vhead:: (Card size, Card size1, CardSum (D1 Sz) size1 size) => 
        (Vec size a) -> Vec (D1 Sz) a
vhead va = listVec (D1 Sz) $ [head (velems va)]

vtail:: (Card size, Card size1, CardSum (D1 Sz) size1 size) => 
        (Vec size a) -> Vec size1 a
vtail va = result
  where result = listVec (vlength_t result) $ tail (velems va)
	

-- Generically sized vectors

newtype VGen a = VGen (Array Int a) deriving Show

tovGen (Vec a) = VGen a

vdlength (VGen a) = let (0,ub) = bounds a in ub + 1

-- Run-time witnessing to a static constraint.
-- The function fromvGen, together with vec and listVec, is a part 
-- of the trusted core. Only those three functions may use
-- the Vec constructor. The trusted core is very small indeed.
-- The function fromvGen (along with num2c below) lets us convert a generic 
-- vector into a statically-sized one -- in a marked contrast with
-- almost any other language.

fromvGen:: (Card c) => VGen a -> c -> Vec c a
fromvGen vg@(VGen a) desired_c 
   = if actual == desired then Vec a
     else error $ "fromvGen: actual " ++ (show actual) ++
                  " desired: " ++ (show desired)
 where
      actual = vdlength vg
      desired = c2num desired_c


-- A slightly different approach to the derivation of CardSum

-- drs = ds1 + ds2 + carry_in
class DigitsSum ds1 ds2 cin dsr | ds1 ds2 cin -> dsr
instance DigitsSum Sz Sz Carry0 Sz
instance DigitsSum Sz Sz Carry1 (D1 Sz)
instance (DigitsSum (D0 Sz) (d2 d2rest) cin (d12 d12rest)) =>
         DigitsSum Sz (d2 d2rest) cin (d12 d12rest)
instance (DigitsSum (d1 d1rest) (D0 Sz) cin (d12 d12rest)) =>
         DigitsSum (d1 d1rest) Sz cin (d12 d12rest)

instance (FullAdder d1 d2 cin d12 cout, 
          DigitsSum d1rest d2rest cout d12rest) =>
   DigitsSum (d1 d1rest) (d2 d2rest) cin (d12 d12rest)

-- drs = ds1 + ds2 + carry_in
-- But now drs is known but ds2 is not
-- carry_in actually means borrow
-- Because our numbers are all non-negative, dsr should have
-- at least as many digits as ds1. Otherwise, the relation is
-- undefined and the subtraction is not possible.
class DigitsDif ds1 ds2 cin dsr | ds1 dsr cin -> ds2
instance DigitsDif Sz ds Carry0 ds
--instance DigitsDif' Sz Sz Carry1 (D1 Sz)
instance (DigitsDif (D0 Sz) (d2 d2rest) Carry1 (d12 d12rest)) =>
         DigitsDif Sz (d2 d2rest) Carry1 (d12 d12rest)

instance (FullAdder d1 d2 cin d12 cout, 
          DigitsDif d1rest d2rest cout d12rest) =>
   DigitsDif (d1 d1rest) (d2 d2rest) cin (d12 d12rest)

-- c12= c1 + c2: c1 and c2 are known.
-- Because c2 is known, we can pattern-match on it.
class CardSum' c1 c2 c12 | c1 c2 -> c12

instance CardSum' c Sz c

instance (DigitsInReverse c1 c1r, 
          DigitsInReverse (dc2 c2) c2r,
	  DigitsSum c1r c2r Carry0 c12r,
	  DigitsInReverse c12r c12)
	  => CardSum' c1 (dc2 c2) c12

{-
-- The following instance, albeit highly desirable, violates the
-- functional dependency: c1 c2 -> c12
instance (CardSum' c2 c4 c2', CardSum' c12 c4 c12', 
          CardSum' c1 c2 c12) => CardSum' c1 c2' c12'
-}

-- c12= c1 + c2: c12 and c1 are known
-- Because c12 is known, we can pattern-match on it. Hence the two
-- instances.
class CardSum'' c1 c2 c12 | c1 c12 -> c2

instance CardSum'' Sz Sz Sz

instance (DigitsInReverse c1 c1r, 
	  DigitsDif c1r c2r' Carry0 c12r,
	  DigitsInReverse c2r' c2', NoLeadingZeros c2' c2,
	  DigitsInReverse c12r (dc12 c12))
	  => CardSum'' c1 c2 (dc12 c12) 


-- Now combine CardSum' and CardSum'' and combine the two
-- dependencies
class CardSum c1 c2 c12 | c1 c2 -> c12, c1 c12 -> c2
	  
instance (CardSum' c1 c2 c12,CardSum'' c1 c2 c12) => CardSum c1 c2 c12

-- An expression of the algebraic law:
-- c1 + c2 = c12, Card c  |- c1 + (c2 + c) = (c12 + c)

instance (Card c, CardSum c1 c2 c12) => CardSum c1 (CardC c c2) (CardC c c12)

card_sum:: (Card c1, Card c2, Card c12, CardSum c1 c2 c12) =>
           c1 -> c2 -> c12
card_sum = undefined


{-
data GenC = forall c. (Card c) => GenC c

-- The following obviously won't do: if CardSum c1 c2 c12 holds, then
-- the run-time check is redundant. If we want a run-time witness,
-- we need to make an instance of the CardSum class!
-- Otherwise, nothing guarantees that the CardSum constraint holds.
data SumCertificate = forall c1 c2 c12. 
                      (Card c1, Card c2, Card c12, CardSum c1 c2 c12) =>
		      SumCertificate c1 c2 c12
    sum_certify (GenC c1) (GenC c2) (GenC c12) | 
                 (c2num c1) + (c2num c2) == (c2num c12) =
                 SumCertificate c1 c2 c12

-- Incidentally, here's an example of a typeof switch
c2d:: GenC -> (forall d. Digits d => d -> w) -> w
c2d (GenC c) f | c2num c == 0 = f Sz
c2d (GenC c) f | c2num c == 1 = f (D1 Sz)


-}		      

-- :t card_sum (D1$ D0$ D1 Sz) (D1 Sz)
-- :t card_sum (D1 Sz) (D1$ D0$ D1 Sz) 
-- :t card_sum (D1$D9 Sz) (D1 Sz)
-- :t card_sum (D1 Sz) (D1$D9 Sz)
-- :t card_sum (D1$D1 Sz) (D1$D9 Sz)
-- :t card_sum (D9$D9 Sz) (D3 Sz)
-- :t card_sum (D1 Sz) (D9$D9 Sz)
-- :t \v -> card_sum (D1$D1 Sz) v `asTypeOf` (D3$D0 Sz)
-- :t \v -> card_sum (D1 Sz) v `asTypeOf` (D1$D0$D0 Sz)
-- :t \v -> card_sum (D9$D9 Sz) v `asTypeOf` (D1$D0$D0 Sz)


-- "Converting" from a dynamic to a static type
-- From an integer to a type representing that particular integer.
-- The latter type is essentially dynamic -- but the constraint
-- Card c is static.

-- rank-2 type: signature is needed.
num2c':: Int -> (forall c. (Card c) => c -> w) -> w
num2c' 0 f = f Sz
num2c' n f  = aux n f Sz
  where
       -- Polymorphic recursion: signature is needed
       aux:: (Digits c) => Int -> (forall c. (Card c) => c -> w) -> c -> w
       aux n f acc = case quotRem n 10 of
                       -- (0,0) may not occur!
                       (0,1)  -> f (D1 acc)
		       (0,2)  -> f (D2 acc)
		       (0,3)  -> f (D3 acc)
		       (0,4)  -> f (D4 acc)
		       (0,5)  -> f (D5 acc)
		       (0,6)  -> f (D6 acc)
		       (0,7)  -> f (D7 acc)
		       (0,8)  -> f (D8 acc)
		       (0,9)  -> f (D9 acc)
		       (nr,0) -> aux nr f (D0 acc)		       
		       (nr,1) -> aux nr f (D1 acc)
		       (nr,2) -> aux nr f (D2 acc)
		       (nr,3) -> aux nr f (D3 acc)
		       (nr,4) -> aux nr f (D4 acc)
		       (nr,5) -> aux nr f (D5 acc)
		       (nr,6) -> aux nr f (D6 acc)
		       (nr,7) -> aux nr f (D7 acc)
		       (nr,8) -> aux nr f (D8 acc)
		       (nr,9) -> aux nr f (D9 acc)

-- Note how this function mirrors the definition of the class Card.
-- In fact, the type system will not let us write the function in any
-- other way! The class Card says that (D0 x) may not be a Card number.
-- Thus the code of the function must _statically_ gurantee that
-- in an application of num2c n f, f will never be applied to anything like
-- (D0 x). That guarantee must be static! And indeed, the above function
-- guarantees that. If we made a mistake and wrote
-- (0,1)  -> f (D0 acc), we would get an error.
-- This is how non-trivial invariants are truly enforced!


-- *Reverse> num2c 12 c2num
-- 12
-- *Reverse> num2c 100 c2num
-- 100

-- However, if the type c is 'dynamic' (generic), we don't
-- actually know the decimal sequence of c. We only know the
-- constraint Card c. Thus when converting an integer to a generic
-- Card number, we merely need to guarantee the constraint. We can
-- do that in a way simpler than constructing a decimal sequence.
-- The following code is obviously faster then the previous one.

-- rank-2 type: signature is needed.
num2c:: Int -> (forall c. (Card c) => c -> w) -> w
num2c 0 f = f Sz
num2c 1 f = f (D1 Sz)
num2c 2 f = f (D2 Sz)
num2c n f  = num2c n1 (\c1 -> num2c n2 (\c2 -> f $ cardc c1 c2))
  where
       n1 = n `div` 2; n2 = n - n1


-- The following generates two generic Card types c and cp1 which are
-- statically known to satisfy the CardSum constraint.

-- Here is the type of a (CPS) consumer of such generic Card types
-- with the (eventual) answer type 'w'
type CC1C w = forall c cp1. (Card c,Card cp1, CardSum (D1 Sz) c cp1) => 
              c -> cp1 -> w

-- Generates the types c and cp1 such as (c2num c) is the same as 
-- the given _run-time_ integer.
-- The second argument is the CPS consumer of the generated types.
num2cn:: Int -> CC1C w -> w
num2cn 0 f = f Sz undefined
num2cn n f = num2c n (\c -> f (cardc c Sz) (cardc c (D1 Sz)))

{-
-- The following didn't work: the reason is that nothing would guarantee
-- the CardSum (D1 Sz) c cp1 constraint of aux on its _recursive_ call.
num2cn n f  = aux_minor n f aux
  where
       aux_minor:: Int -> CC1C w -> (forall c cp1 p. (Card c,Card cp1, CardSum (D1 Sz) c cp1, Digits p) => c -> cp1 -> p -> Int -> CC1C w -> w) -> w
       aux_minor n f g = case m of
                        0 -> g     Sz  undefined (D0 Sz) nr f
                        1 -> g (D1 Sz) undefined (D0 Sz) nr f -- etc.
	   where (nr,m) = quotRem n 10
       -- Polymorphic recursion: signature is needed
       aux:: (Card c,Card cp1, CardSum (D1 Sz) c cp1, Digits p) => c -> cp1 -> p -> Int -> CC1C w -> w
       aux c cp1 p _ n f = case quotRem n 10 of
                       (0,0) ->  f c cp1 
                       (n,1)  -> aux (card_sum (D1 p) c) -- recursive call!
		                     undefined (D0 p) undefined n f -- etc.
-}

-- This code exposed a bug in GHC prior to 6.2.2
-- Message on glasgow-haskell-bugs@haskell.org posted on Feb 4, 2004.

-- some tests
testnh n = num2cn n (\ (c1::c1) (c::c) -> let (v::Vec c Char) = Vec (listArray (0,12) ['a']) in vhead v )
testnt n = num2cn n (\ (c1::c1) (c::c) -> let (v::Vec c Char) = Vec (listArray (0,12) ['a']) in vlength $ vtail v)



-- Finally, the desired function. Obviously, the size of the result
-- is equal to the size of the argument vector.
-- In the code below, vhead', and vtail' are _statically_ guaranteed
-- to be applied to a non-empty vector. Therefore, we can _safely_ use
-- the versions of vhead' and vtail' that do no run-time size checks.
vreverse v = fromvGen (aux (tovGen v) (tovGen (vec Sz undefined)))
                      (vlength_t v)
  where
    aux v acc | vdlength v == 0 = acc
    aux v acc = num2cn ((vdlength v)-1) f
      where
	f c1 c = let vs = fromvGen v c
	             g ac_c acc_p1 = 
		         let sacc = fromvGen acc ac_c
		         in aux (tovGen (vtail vs))
 			        (tovGen (vappend (vhead vs) sacc))
                 in num2cn (vdlength acc) g


-- vreverse v12i
-- vreverse v13i
{-
-- The following naive code does not type: vhead requires a _static_
-- guarantee that it will be given a non-zero vector
vreverse v = aux v (vec Sz undefined)
  where
    aux:: (Card size1, Card size2, Card size3,
           CardSum size1 size2 size3) => 
	   (Vec size1 a) -> (Vec size2 a) -> (Vec size3 a)
    aux v acc | vlength v == 0 = acc
    aux v acc = aux (vtail v) (vappend (vhead v) acc)
-}
{--
-- The following, more sophisticated attempt doesn't type either:
-- Just because the relation 
-- VRevAux (Vec (dv vsize) a) (Vec acsize a) (Vec rsize a)
-- holds, does not mean that 
-- VRevAux vsize1 acsize1 rsize
-- will hold too. The latter is necessary to type the recursive
-- call to vrevaux

class VRevAux v acc result | v acc -> result where
    vrevaux:: v -> acc -> result
    
instance VRevAux (Vec Sz a) acc acc where
    vrevaux v acc = acc
    
instance (Card vsize1, Card (dv vsize),
          Card rsize, Card acsize, Card acsize1,
          CardSum (D1 Sz) acsize acsize1,
          VRevAux vsize1 acsize1 rsize, 
	  CardSum (dv vsize) acsize rsize,
          CardSum (D1 Sz) vsize1 (dv vsize)) =>
	 VRevAux (Vec (dv vsize) a) (Vec acsize a) (Vec rsize a) where
    vrevaux v acc = 
        --vappend v acc
          vrevaux (vtail v) 
	          (vappend (vhead v) acc)
--}
			    
