{-# OPTIONS -fglasgow-exts -fallow-undecidable-instances  #-}

-- Number-parameterized vectors: unary numbers, data-constructor-based
-- encoding

module UnaryDS where

data VZero a = VZero deriving Show

infixr 3 :+:
data Vecp tail a = a :+: (tail a) deriving Show

class Vec t where
    vlength::  t a -> Int
    vat::      t a -> Int -> a
    vzipWith:: (a->b->c) -> t a -> t b -> t c
   
instance Vec VZero where
    vlength  = const 0
    vat      = error "null array or index out of range"
    vzipWith f a b = VZero

instance (Vec tail) => Vec (Vecp tail) where
    vlength (_ :+: t) = 1 + vlength t
    vat (a :+: _)  0 = a
    vat (_ :+: ta) n = vat ta (n-1)
    vzipWith f (a :+: ta) (b :+: tb) = (f a b) :+: (vzipWith f ta tb)

v3c = 'a' :+: 'b' :+: 'c' :+: VZero
v3i = 1 :+: 2 :+: 3 :+: VZero
v4i = 1 :+: 2 :+: 3 :+: 4 :+: VZero

-- :t v3c
-- Note from the inferred type the separation of the skeleton and the content

test1 = vlength v3c
test2 = [vat v3c 0, vat v3c 1, vat v3c 2]

test3 = vzipWith (\c i -> (toEnum $ fromEnum c + fromIntegral i)::Char) v3c v3i
test4 = vzipWith (+) v3i v3i

-- The following leads to a type error, as expected:
-- test5 = vzipWith (+) v3i v4i

